import numpy as np
import pandas as pd
from sklearn.feature_selection import f_regression


def softmax(x, tau = 1):
    if isinstance(x, list):
        return np.exp([y/tau for y in x])/sum(np.exp([y/tau for y in x])) 
    return np.exp(x/tau)/sum(np.exp(x/tau))


def calc_f_score(X, y, sort = True):
    f = pd.Series(f_regression(X, y)[0], index=X.columns)
    if sort:
        f = f.sort_values(ascending=False)
    return f



from genetic_feature_selection.genetic_funcs import _replace_dup_genes, mate_parents, evaluate_generation, weighted_selection
import numpy as np
from genetic_feature_selection.genetic_stats import softmax


def roulette(
    evaluted_pop, num_parents_mating, n_genes, fitness_func
):
    probs = softmax([i[0] for i in evaluted_pop])
    parents = [i[1] for i in evaluted_pop]
    children = []

    # make mating group 
    # this is done by selecting num_parents_mating 
    # indices, these will be parents.
    mates = np.random.choice(
        len(parents),
        size=num_parents_mating, 
        replace=False, 
        p=probs
    )

    parents2mate = []
    for i in mates:
        parents2mate.append(evaluted_pop[i])
    parents2mate.sort(reverse=True, key=lambda x: x[0])

    # use indices to split mating group in to list 
    # of pairs [[p1, p2], [p3, p4] ...]
    mating_pairs = []
    i = 0 
    while (i < len(mates)):
        pair = []
        for _ in range(2):
            pair.append(parents2mate[i])
            i += 1
        mating_pairs.append(pair)

    # produce children from mating pairs
    for pair in mating_pairs:
        # child = mate_parents(pair[0][1], pair[1][1])
        child = weighted_selection(pair[0], pair[1])
        child = _replace_dup_genes(child, n_genes)
        children.append(child)

    evaluted_children = evaluate_generation(
        children, fitness_func
    )
    # evaluted_children.sort(reverse=True, key=lambda x: x[0])

    return evaluted_children





def main():
    p = [
        (0.9, [1, 2, 1, 4]), 
        (0.88, [5, 6, 1, 8]),
        (0.87, [1, 10, 1, 12]),
        (0.84, [1, 14, 1, 16]),
        (0.77, [1, 18, 1, 20]),
        (0.50, [1, 22, 1, 24])
    ]

    def fitness_func(a,b):
        return sum(a)

    print(roulette(p, 4, 24, fitness_func))


if __name__ == "__main__":
    main()



import warnings
warnings.filterwarnings("ignore")
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import average_precision_score
from genetic_feature_selection.genetic_funcs import get_features
from sklearn.model_selection import train_test_split
from genetic_feature_selection.genetic_search import GeneticSearch
from genetic_feature_selection.f_score_generator import FScoreSoftmaxInitPop
import pandas as pd
import pytest

@pytest.fixture
def X_y():
    X, y = make_classification(n_samples=1000, n_informative=20, n_redundant=0)
    X = pd.DataFrame(X)
    X.columns = [f"col_{i}" for i in X.columns]
    y = pd.Series(y, name="y", dtype=int)
    X_train, X_test, y_train, y_test = train_test_split(X, y)
    return X_train, X_test, y_train, y_test


def test_search(X_y):
    X_train, X_test, y_train, y_test = X_y

    gsf = GeneticSearch(
        iterations = 5, 
        sol_per_pop = 4, 
        generations = 15, 
        keep_n_best_individuals = 4, 
        select_n_features = 5,
        num_parents_mating = 4,
        X_train = X_train,
        y_train = y_train,
        X_test = X_test,
        y_test = y_test,
        clf = LogisticRegression,
        clf_params = dict(max_iter=15),
        probas = True,
        scorer = average_precision_score,
        gen_pop = FScoreSoftmaxInitPop(
            X_train, y_train, tau = 50
        )
    )



    best_cols = gsf.search()

    assert isinstance(best_cols[0], str)
    assert best_cols[0].startswith("col_")
    assert len(best_cols) == 5



import numpy as np
from genetic_feature_selection.genetic_stats import softmax, calc_f_score
from abc import ABC, abstractmethod


class FScoreInitPop(ABC):
    def __init__(self, X, y) -> None:
        self.f_scores = calc_f_score(X, y)
        self.y = y
        self.map_col_to_index(X)
        self.X = X.rename(columns = self.col2int)

    @abstractmethod
    def gen_pop(self):
        pass

    def map_col_to_index(self, X):
        self.int2col = {i:col for i,col in enumerate(X.columns)}
        self.col2int = {col:i for i,col in enumerate(X.columns)}

    def int2str(self, best_cols):
        return [self.int2col[c] for c in best_cols]


class GreedyFScoreInitPop(FScoreInitPop):
    def gen_pop(self, soln_per_pop, n_features, select_n_features):
        # f_scores <- calculate f-score 
        # generate i empty feature vectors
        # for each x_i in f_score:
        #   add x_i to to feature_vec_i in such a manner that
        #   feature_vec_i = [x_i, x_(2*i), ..., x_(n_features*i)]
        i = 0
        j = 0
        solns = []

        for _ in range(select_n_features):
            for _ in range(soln_per_pop):
                if len(solns) < soln_per_pop:
                    solns.append([])
                solns[i].append(self.col2int[self.f_scores.index[j]])
                
                i += 1
                if i >= self.soln_per_pop:
                    i = 0
                j += 1

        return solns


class FScoreSoftmaxInitPop(FScoreInitPop):
    def __init__(self, X, y, tau = 50) -> None:
        super().__init__(X, y)
        self.f_scores = calc_f_score(X, y, sort = False).fillna(0)
        self.tau = tau
        self.probs = softmax(self.f_scores, self.tau)

        if any(np.isnan(self.probs)):
            raise ValueError(
                "Det er NaNs i sannsynlighetene. Dette kan bety at noen variabler har fått svært høy f-score. Prøv å øke verdien til tau parameteren i ")

    def gen_pop(self, soln_per_pop, n_features, select_n_features):
        selected_features = np.random.choice(
            self.X.columns, size = (select_n_features, soln_per_pop),
            p = self.probs, replace=False
        )
        
        return selected_features.T.tolist()



import numpy as np
from genetic_feature_selection.genetic_stats import softmax
import random


def get_features(X, genes):
    return X.iloc[:,genes]


def evaluate_generation(population, fitness_func):
    evaluted_gen = []

    for ind in population:
        score = fitness_func(ind)
        evaluted_gen.append((score, ind))
    return evaluted_gen


def mate_parents(parent1, parent2):
    parent1, parent2 = list(parent1), list(parent2)
    child = parent1[:int(len(parent1)/2)] + parent2[int(len(parent1)/2):]
    assert len(child) == len(parent1)
    return child


def weighted_selection(p1, p2):
    probs = softmax([p1[0], p2[0]])
    select_from = np.random.choice([0,1], size=len(p1[1]), p=probs).tolist()
    child = []
    parents = (p1, p2)

    for i,p in enumerate(select_from):
        child.append(parents[p][1][i])

    return child


def _add_random_gene(ind, desired_len, max_gene):
    if len(ind) == desired_len:
        return ind
    
    rand_int = random.randrange(max_gene)
    if rand_int not in ind:
        ind.append(rand_int)
    
    return _add_random_gene(ind, desired_len, max_gene)


def _replace_dup_genes(child, n_genes):
    # ertstatt dupliserte gener med tilfeldige gener som ikke eksisterer i individet 
    no_dups = set(child)

    if len(no_dups) < len(child):
        child_w_dups = list(no_dups)

        child = _add_random_gene(child_w_dups, len(child), n_genes)

    return child


def main():
    efvs = [
        (0.9, [1,2,3,4]),
        (0.0, [5,6,7,8]),
    ]

    p1, p2 = efvs[0], efvs[1]

    # child = weighted_selection(
    #     efvs[0], 
    #     efvs[1]
    # )

    # print(child)

    probs = softmax([p1[0], p2[0]])
    print(probs)
    select_from = np.random.choice([0,1], size=len(p1[1]), p=probs).tolist()
    child = []
    parents = (p1, p2)

    for i,p in enumerate(select_from):
        child.append(parents[p][1][i])

if __name__ == "__main__":
    main()



import pathlib
from typing import Callable, Dict, Any
from genetic_feature_selection.genetic_funcs import *
from genetic_feature_selection.roulette_wheel_selection import roulette
from genetic_feature_selection.vanilla_generator import GenPopVanilla
from genetic_feature_selection.genetic_iter import GeneticIter
import pandas as pd


class GeneticSearch:
    """Algoritme for genetisk variabel søk. 

    For hver iterasjon genereres det generations antall generasjoner med sol_per_pop antall
    individer. En generasjon er en runde med generering, evaluering og videreføring av de 
    keep_n_best_individuals individene til neste generasjon. 
    Et individ er en vektor med variabler. Hvert genererte individ vil bli gitt en score. Denne
    scoren er med på vekte sannsynligheten for at dette individet for parre seg. Scoren bestemmer
    også hvilke individer som vil bli med til neste generasjon. 
    Scoren til hvert individ kalkuleres ved hjelp av fitness_func. fitness_func er en funksjon som
    trener en modell med variablene angitt i individet og evaluerer denne modellen med en angitt 
    metrikk. 

    Args:
        iterations (int): Hvor mange iterasjoner skal gjennomføres. Det vil bli trent 
            iterations * generations * sol_per_pop modeller når metoden train kalles. 
        sol_per_pop (int): Hvor mange feature vectors skal genereres for hver generasjon. 
        generations (int): Hvor mange generasjoner per iterasjon. 
        n_features (int): Hvor mange features eksisterer det i datasettet.
        keep_n_best_individuals (int): Hvor mange av de beste individene skal videre med til
            neste generasjon.
        select_n_features (int): Hvor mange variabler skal velges. 
        fitness_func (Callable[[List[int]], float]): Funksjon som regner ut score for hvert individ.
        num_parents_mating (int): Hvor mange av individene skal pare seg. 
        gen_pop (_type_, optional): Hvordan skal nye individer genereres Denne klassen må implementere
            en gen_pop metode for å generere en ny generasjon med individer. Defaults to GenPopVanilla().
    """
    def __init__(
        self,
        iterations : int,
        sol_per_pop : int,
        generations : int,
        keep_n_best_individuals : int,
        select_n_features : int,
        num_parents_mating : int,
        X_train : pd.DataFrame,
        y_train : pd.Series,
        X_test : pd.DataFrame,
        y_test : pd.Series,
        clf : Any,
        probas : bool, 
        scorer : Callable[[pd.Series, pd.Series], float], 
        gen_pop = GenPopVanilla(),
        path : str = None,
        clf_params : Dict[str, Any] = {},
        fitness_func = None
    ) -> None:
        self.iterations = iterations
        self.sol_per_pop = sol_per_pop
        self.generations = generations
        self.n_features = X_train.shape[1]
        self.keep_n_best_individuals = keep_n_best_individuals
        self.select_n_features = select_n_features
        self.num_parents_mating = num_parents_mating
        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test
        self.clf = clf
        self.clf_params = clf_params
        self.probas = probas
        self.scorer = scorer
        self.pop_generator = gen_pop
        self.gen_pop = gen_pop.gen_pop
        self.path = path
        if fitness_func: setattr(self, "fitness_func", fitness_func)
        self.int2col = {i:col for i,col in enumerate(X_train.columns)}
        self.col2int = {col:i for i,col in enumerate(X_train.columns)}


    def _dump_results(self, txt : str):
        if not self.path:
            self.path = pathlib.Path("search-log.log")
        else:
            self.path = pathlib.Path(self.path)
        
        with open(self.path, "a") as f:
            f.write(txt + "\n")
        
    def search(self, dump = True):
        self.results = []

        for i in range(self.iterations):
            print(f"Iteration {i}")
            best_cols, score, iter_scores = self.one_iteration()
            self.results.append(GeneticIter(i, score, best_cols, iter_scores))

            if dump:
                self._dump_results(str(self.results[-1]))

        self.results.sort(key=lambda x: x.score)
        return self.results[-1].cols
    
    def one_iteration(self):
        pop = self.gen_pop(
            self.sol_per_pop, self.n_features, self.select_n_features)
        prev_best = 0
        scores = []
        keep = []

        for i in range(self.generations):
            evaluted_pop = evaluate_generation(pop, self.fitness_func)

            # Check if child is better than the second best individual in the population
            # This is done to avoid sorting the list twice 
            evaluated_children = roulette(
                evaluted_pop, self.num_parents_mating, self.n_features, self.fitness_func
            )

            evaluted_pop.extend(evaluated_children)
            evaluted_pop.sort(reverse=True, key=lambda x: x[0])

            # keep keep_n_best_individuals best individuals and create a new generation.  
            keep = evaluted_pop[:self.keep_n_best_individuals]
            keep = [i[1] for i in keep]
            
            # add crossover gen to pop list if exists, this is done so we dont have to reevaluate the individuals
            pop = keep + self.gen_pop(
                self.sol_per_pop-self.keep_n_best_individuals, self.n_features, self.select_n_features
            )
            
            curr_best_score = evaluted_pop[0][0]
            scores.append(curr_best_score)

            if curr_best_score > prev_best:
                print("\t generation:", i, "score:", curr_best_score)
                prev_best = curr_best_score
                best_cols = evaluted_pop[0][1]
        
        best_cols = self.int2str(best_cols)
        return best_cols, prev_best, scores

    def fitness_func(self, soln):    
        X_train_soln = get_features(self.X_train, soln)
        X_val_son = get_features(self.X_test, soln)

        clf = self.clf(**self.clf_params)
        clf.fit(X_train_soln, self.y_train)

        if self.probas:
            preds = clf.predict_proba(X_val_son)[:,1]
        else:
            preds = clf.predict(X_val_son)[:,1]

        return self.scorer(self.y_test, preds)

    def int2str(self, best_cols):
        return [self.int2col[c] for c in best_cols]



from setuptools import find_packages
from setuptools import setup
from __version__ import __version__


def read(file_name):
    with open(file_name, "r") as f:
        txt = f.read()
    return txt


setup(
    name='genetic_feature_selection',
    description='Program for selecting the n best features for your machine learning model.',
    long_description=read("README.rst"),
    author='Magnus P. Nytun',
    author_email='magnus.nytun@sb1ostlandet.no',
    version=__version__,
    packages= find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3.8',
    ],
    install_requires=[
        "numpy",
        "pandas",
        "scikit-learn"
    ]
)
